def readable_seconds(s: float):
    minutes = int(s / 60)
    seconds = int(s % 60)
    return "".join([
        f"{minutes} minute{'s' if minutes > 1 else ''}{' and ' if seconds > 0 else ''}" if minutes > 0 else "",
        f"{seconds} second{'s' if seconds > 1 else ''}" if seconds > 0 else ""
    ])



from datetime import datetime
from typing import Union, Dict, Any, Optional

import requests

from .config import Class
from .consts import ADD_BOOKING_URL, CLASSES_SCHEDULE_URL, WEEKDAYS, CANCEL_BOOKING_URL, CLASS_URL
from .errors import BookingError


def book_class(token, class_id) -> bool:
    print(f"[INFO] Booking class {class_id}")
    response = requests.post(
        ADD_BOOKING_URL,
        {
            "classId": class_id,
            "token": token
        }
    )
    if response.status_code != requests.codes.OK:
        print("[ERROR] Booking attempt failed: " + response.text)
        return False
    return True


def cancel_booking(token, class_id) -> bool:
    print(f"[INFO] Cancelling booking of class {class_id}")
    res = requests.post(
        CANCEL_BOOKING_URL,
        {
            "classId": class_id,
            "token": token
        }
    )
    if res.status_code != requests.codes.OK:
        print("[ERROR] Booking cancellation attempt failed: " + res.text)
        return False
    body = res.json()
    if body["success"] is False:
        print("[ERROR] Booking cancellation attempt failed: " + body.errorMessage)
        return False
    if "class" not in body or "userStatus" not in body["class"] or body["class"]["userStatus"] != "available":
        print("[ERROR] Booking cancellation attempt failed, class is still booked!")
        return False
    return True


# Search the scheduled classes and return the first class matching the given arguments
def find_class(token: str, _class_config: Class) -> Union[Dict[str, Any], BookingError]:
    print(f"[INFO] Searching for class matching config: {_class_config}")
    schedule_response = requests.get(
        f"{CLASSES_SCHEDULE_URL}?token={token}&studios={_class_config.studio}&lang=no"
    )
    if schedule_response.status_code != requests.codes.OK:
        print("[ERROR] Schedule get request denied")
        return BookingError.ERROR
    schedule = schedule_response.json()
    if 'days' not in schedule:
        print("[ERROR] Malformed schedule, contains no days")
        return BookingError.MALFORMED_SCHEDULE
    days = schedule['days']
    target_day = None
    if not 0 <= _class_config.weekday < len(WEEKDAYS):
        print(f"[ERROR] Invalid weekday number ({_class_config.weekday=})")
        return BookingError.MALFORMED_SEARCH
    weekday_str = WEEKDAYS[_class_config.weekday]
    for day in days:
        if 'dayName' in day and day['dayName'] == weekday_str:
            target_day = day
            break
    if target_day is None:
        print(f"[ERROR] Could not find requested day '{weekday_str}'. To early?")
        return BookingError.MISSING_SCHEDULE_DAY
    classes = target_day['classes']
    result = None
    for c in classes:
        if 'activityId' not in c or c['activityId'] != _class_config.activity:
            continue
        start_time = datetime.strptime(c['from'], '%Y-%m-%d %H:%M:%S')
        time_matches = start_time.hour == _class_config.time.hour and start_time.minute == _class_config.time.minute
        if not time_matches:
            print(f"[INFO] Found class, but start time did not match: {c}")
            result = BookingError.INCORRECT_START_TIME
            continue
        if 'name' not in c:
            print(f"[WARNING] Found class, but the name is missing: {c}")
            result = BookingError.MALFORMED_CLASS
            continue
        search_feedback = f"[INFO] Found class: \"{c['name']}\""
        if 'instructors' in c and len(c['instructors']) > 0 and 'name' in c['instructors'][0]:
            search_feedback += f" with {c['instructors'][0]['name']}"
        else:
            search_feedback += " (missing instructor)"
        if 'from' in c:
            search_feedback += f" at {c['from']}"
        print(search_feedback)
        return c
    print("[ERROR] Could not find class matching criteria")
    if result is None:
        result = BookingError.CLASS_MISSING
    return result


def find_class_by_id(token: str, class_id: str) -> Optional[Dict[str, Any]]:
    print(f"[INFO] Searching for class by id: {class_id}")
    class_response = requests.get(
        f"{CLASS_URL}?token={token}&id={class_id}&lang=no"
    )
    if class_response.status_code != requests.codes.OK:
        print("[ERROR] Class get request failed")
        return None
    return class_response.json()["class"]



import datetime

from sit_rezervo.config import Class, Cron

STARTUP_DAYS_BEFORE_ACTIVITY = 2


def generate_booking_cron_job(index: int, _class: Class, cron_config: Cron, config_path: str) -> str:
    # Using current datetime simply as a starting point
    # We really only care about the "wall clock" part, which is replaced by input values
    activity_time = datetime.datetime.now().replace(
        hour=_class.time.hour,
        minute=_class.time.minute,
        second=0, microsecond=0  # Cosmetic only
    )
    # Back up time to give booking script some prep time
    booking_time = activity_time - datetime.timedelta(minutes=cron_config.preparation_minutes)
    # Handle case where backing up time changes the weekday
    booking_weekday_delta = activity_time.weekday() - booking_time.weekday()
    booking_weekday = (_class.weekday + 1 - STARTUP_DAYS_BEFORE_ACTIVITY - booking_weekday_delta) % 7
    program_command = f"cd {cron_config.sit_rezervo_dir} || exit 1; PATH=$PATH:/usr/local/bin " \
                      f"{cron_config.python_path}/sit-rezervo book -c {config_path}"
    output_redirection = f">> {cron_config.log_path} 2>&1"
    class_name = _class.display_name if _class.display_name is not None else _class.activity
    booking_cron_job = (
        f"# {class_name}\n"
        f"{booking_time.minute} {booking_time.hour} * * {booking_weekday} "
        f"{program_command} {index} "
        f"{output_redirection}"
        "\n"
    )
    if cron_config.precheck_hours is not None:
        precheck_time = booking_time - datetime.timedelta(hours=cron_config.precheck_hours)
        precheck_weekday = booking_weekday
        precheck_cron_job = (
            f"# {class_name} (precheck)\n"
            f"{precheck_time.minute} {precheck_time.hour} * * {precheck_weekday} "
            f"{program_command} {index} --check "
            f"{output_redirection}"
            "\n"
        )
        return f"{precheck_cron_job}{booking_cron_job}"
    return booking_cron_job



from enum import Enum, auto


class BookingError(Enum):
    ERROR = auto()
    MALFORMED_SEARCH = auto()
    MALFORMED_SCHEDULE = auto()
    MISSING_SCHEDULE_DAY = auto()
    INCORRECT_START_TIME = auto()
    CLASS_MISSING = auto()
    MALFORMED_CLASS = auto()
    TOO_LONG_WAITING_TIME = auto()
    INVALID_CONFIG = auto()
    CANCELLING_WITHOUT_BOOKING = auto()



import requests


def upload_ical_to_transfersh(transfersh_url: str, ical_url: str, filename: str) -> str:
    return transfersh_direct_url(
        requests.post(transfersh_url, files={filename: requests.get(ical_url).text}).text
    )


def transfersh_direct_url(transfersh_url: str):
    url_parts = transfersh_url.split("/")
    return "/".join(url_parts[:3]) + "/get/" + "/".join(url_parts[3:])



from dataclasses import dataclass
from io import TextIOWrapper
from typing import Optional

from dataclass_wizard import YAMLWizard


@dataclass
class Auth:
    email: str
    password: str
    max_attempts: int = 3


@dataclass
class Booking:
    timezone: str
    max_attempts: int = 10
    max_waiting_minutes: int = 60


@dataclass
class Cron:
    sit_rezervo_dir: str
    python_path: str
    log_path: str
    preparation_minutes: int = 10
    precheck_hours: Optional[int] = None


@dataclass
class Transfersh:
    url: str


@dataclass
class Slack:
    bot_token: str
    channel_id: str
    user_id: str


@dataclass
class Notifications:
    transfersh: Optional[Transfersh] = None
    slack: Optional[Slack] = None
    reminder_hours_before: Optional[int] = None


@dataclass
class ClassTime:
    hour: int
    minute: int


@dataclass
class Class:
    activity: int
    weekday: int
    studio: int
    time: ClassTime
    display_name: Optional[str] = None


@dataclass
class Config(YAMLWizard):
    auth: Auth
    booking: Booking
    cron: Cron
    classes: list[Class]
    notifications: Optional[Notifications] = None


def config_from_stream(stream: TextIOWrapper) -> Optional[Config]:
    return Config.from_yaml(stream)



from typing import Dict, Any, Optional

from ..auth import AuthenticationError
from ..config import Notifications, Class
from ..errors import BookingError
from .slack import notify_auth_failure_slack, notify_booking_failure_slack, notify_booking_slack, \
    schedule_class_reminder_slack


def notify_auth_failure(notifications_config: Notifications, error: AuthenticationError = None,
                        check_run: bool = False) -> None:
    slack_config = notifications_config.slack
    if slack_config is not None:
        return notify_auth_failure_slack(slack_config.bot_token, slack_config.channel_id, slack_config.user_id, error,
                                         check_run=check_run)
    print("[WARNING] No notification targets, auth failure notification will not be sent!")


def notify_booking_failure(notifications_config: Notifications, _class_config: Class, error: BookingError = None,
                           check_run: bool = False) -> None:
    slack_config = notifications_config.slack
    if slack_config is not None:
        return notify_booking_failure_slack(slack_config.bot_token, slack_config.channel_id,
                                            slack_config.user_id, _class_config, error, check_run)
    print("[WARNING] No notification targets, booking failure notification will not be sent!")


def notify_booking(notifications_config: Notifications, booked_class: Dict[str, Any], ical_url: str) -> None:
    slack_config = notifications_config.slack
    if slack_config is not None:
        scheduled_reminder_id = None
        if notifications_config.reminder_hours_before is not None:
            scheduled_reminder_id = schedule_class_reminder(notifications_config, booked_class)
        if notifications_config.transfersh is not None:
            transfersh_url = notifications_config.transfersh.url
        else:
            transfersh_url = None
        return notify_booking_slack(slack_config.bot_token, slack_config.channel_id, slack_config.user_id, booked_class,
                                    ical_url, transfersh_url, scheduled_reminder_id)
    print("[WARNING] No notification targets, booking notification will not be sent!")


def schedule_class_reminder(notifications_config: Notifications, booked_class: Dict[str, Any]) -> Optional[str]:
    slack_config = notifications_config.slack
    if slack_config is not None:
        return schedule_class_reminder_slack(slack_config.bot_token, slack_config.user_id,
                                             booked_class, notifications_config.reminder_hours_before)
    print("[WARNING] No notification targets, class reminder will not be sent!")



from pathlib import Path

# Absolute normalized path to the root directory of the app
APP_ROOT = Path(__file__).parent.resolve()

CONFIG_PATH = "config.yaml"

WEEKDAYS = ["Mandag", "Tirsdag", "Onsdag", "Torsdag", "Fredag", "Lørdag", "Søndag"]

AUTH_URL = "https://www.sit.no/"
BOOKING_URL = "https://www.sit.no/trening/gruppe"

ADD_BOOKING_URL = "https://ibooking.sit.no/webapp/api//Schedule/addBooking"
CANCEL_BOOKING_URL = "https://ibooking.sit.no/webapp/api//Schedule/cancelBooking"
CLASSES_SCHEDULE_URL = 'https://ibooking.sit.no/webapp/api/Schedule/getSchedule'
CLASS_URL = 'https://ibooking.sit.no/webapp/api/Schedule/getClass'
TOKEN_VALIDATION_URL = "https://ibooking.sit.no/webapp/api/User/validateToken"
ICAL_URL = "https://ibooking.sit.no/webapp/api/Schedule/calendar"

SLACK_ACTION_ADD_BOOKING_TO_CALENDAR = "add_booking_to_calendar"
SLACK_ACTION_CANCEL_BOOKING = "cancel_booking"



import json
from typing import List, Optional

import pydantic
from fastapi import FastAPI, status, Response, Form, BackgroundTasks, Depends
from pydantic import BaseModel

from .auth import AuthenticationError
from .booking import find_class_by_id
from .config import Config
from .consts import SLACK_ACTION_ADD_BOOKING_TO_CALENDAR, SLACK_ACTION_CANCEL_BOOKING
from .errors import BookingError
from .main import try_cancel_booking, try_authenticate
from .notify.slack import notify_cancellation_slack, notify_working_slack, \
    notify_cancellation_failure_slack, show_unauthorized_action_modal_slack

api = FastAPI()


def get_configs() -> Optional[list[Config]]:
    return None  # will be overriden somewhere...


class User(BaseModel):
    id: str


class Action(BaseModel):
    action_id: str
    value: Optional[str]


class CancelBookingActionValue(BaseModel):
    class_id: str
    user_id: str


class InteractionContainer(BaseModel):
    type: str
    message_ts: str


class Interaction(BaseModel):
    type: str
    trigger_id: str
    user: User
    actions: List[Action]
    response_url: str
    container: InteractionContainer


def find_config_by_slack_id(configs: list[Config], user_id: str) -> Optional[Config]:
    for config in configs:
        if config.notifications is None:
            continue
        if config.notifications.slack is None:
            continue
        if config.notifications.slack.user_id == user_id:
            return config
    return None


def deserialize_cancel_booking_action_value(val: str) -> CancelBookingActionValue:
    as_dict = json.loads(val)
    return CancelBookingActionValue(
        user_id=as_dict['userId'],
        class_id=as_dict['classId']
    )


def handle_cancel_booking_slack_action(config: Config, action_value: CancelBookingActionValue, message_ts: str,
                                       response_url: str):
    if config.notifications is None:
        print("[WARNING] Notifications config not specified, no notifications will be sent!")
        slack_config = None
    else:
        slack_config = config.notifications.slack
        if slack_config is None:
            print("[WARNING] Slack notifications config not specified, no Slack notifications will be sent!")
        else:
            notify_working_slack(slack_config.bot_token, slack_config.channel_id, message_ts)
    print("[INFO] Authenticating...")
    auth_result = try_authenticate(config.auth.email, config.auth.password, config.auth.max_attempts)
    if isinstance(auth_result, AuthenticationError):
        print("[ERROR] Authentication failed, abort!")
        if slack_config is not None:
            notify_cancellation_failure_slack(slack_config.bot_token, slack_config.channel_id, message_ts, auth_result)
        return
    _class = find_class_by_id(auth_result, action_value.class_id)
    if _class is None:
        print("[ERROR] Class retrieval by id failed, abort!")
        if slack_config is not None:
            notify_cancellation_failure_slack(slack_config.bot_token, slack_config.channel_id, message_ts,
                                              BookingError.CLASS_MISSING)
        return
    cancellation_error = try_cancel_booking(auth_result, _class, config.booking.max_attempts)
    if cancellation_error is not None:
        if slack_config is not None:
            notify_cancellation_failure_slack(slack_config.bot_token, slack_config.channel_id, message_ts,
                                              cancellation_error)
        return
    if slack_config is not None:
        notify_cancellation_slack(slack_config.bot_token, slack_config.channel_id, message_ts, response_url)


@api.post("/")
def slack_action(background_tasks: BackgroundTasks, payload: str = Form(),
                 configs: Optional[list[Config]] = Depends(get_configs)):
    if configs is None:
        print("[ERROR] No configs available, abort!")
        return Response(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)
    # TODO: Verify that request is from Slack (https://api.slack.com/authentication/verifying-requests-from-slack)
    interaction: Interaction = pydantic.parse_raw_as(type_=Interaction, b=payload)
    if interaction.type != "block_actions":
        return Response(f"Unsupported interaction type '{interaction.type}'", status_code=status.HTTP_400_BAD_REQUEST)
    if len(interaction.actions) != 1:
        return Response(f"Unsupported number of interaction actions", status_code=status.HTTP_400_BAD_REQUEST)
    action = interaction.actions[0]
    action_value = deserialize_cancel_booking_action_value(action.value)
    if action.action_id == SLACK_ACTION_ADD_BOOKING_TO_CALENDAR:
        return Response(status_code=status.HTTP_200_OK)
    if action.action_id == SLACK_ACTION_CANCEL_BOOKING:
        message_ts = interaction.container.message_ts
        config = find_config_by_slack_id(configs, action_value.user_id)
        if config is None:
            print("[ERROR] Could not find config for Slack user, abort!")
            return Response(status_code=status.HTTP_400_BAD_REQUEST)
        # This check should be performed before retrieving config, but then we wouldn't be able to display a funny modal
        if action_value.user_id != interaction.user.id:
            print("[WARNING] Detected cancellation attempt by an unauthorized user")
            if config.notifications is not None and config.notifications.slack is not None:
                background_tasks.add_task(show_unauthorized_action_modal_slack, config.notifications.slack.bot_token,
                                          interaction.trigger_id)
            return Response("Nice try 👏", status_code=status.HTTP_403_FORBIDDEN)
        background_tasks.add_task(handle_cancel_booking_slack_action, config, action_value, message_ts,
                                  interaction.response_url)
        return Response(status_code=status.HTTP_200_OK)
    return Response(f"Unsupported interaction action", status_code=status.HTTP_400_BAD_REQUEST)



import datetime
import json
import time
from typing import Dict, Any, Optional, List
from requests import RequestException
from slack_sdk import WebClient as SlackClient
from slack_sdk.webhook import WebhookClient as SlackWebhookClient
from slack_sdk.errors import SlackApiError

from ..auth import AuthenticationError
from ..config import Class
from ..consts import WEEKDAYS, SLACK_ACTION_ADD_BOOKING_TO_CALENDAR, SLACK_ACTION_CANCEL_BOOKING
from ..errors import BookingError
from .utils import upload_ical_to_transfersh


def notify_slack(slack_token: str, channel: str, message: str, message_blocks: Optional[List[Dict[str, Any]]] = None,
                 thread_ts: Optional[str] = None):
    try:
        SlackClient(token=slack_token).chat_postMessage(
            channel=channel,
            text=message,
            blocks=message_blocks,
            thread_ts=thread_ts
        )
    except SlackApiError as e:
        print(f"[FAILED] Could not post notification to Slack: {e.response['error']}")
        return False
    return True


def schedule_dm_slack(slack_token: str, user_id: str, post_at: int, message: str,
                      message_blocks: Optional[List[Dict[str, Any]]] = None) -> Optional[str]:
    try:
        res = SlackClient(token=slack_token).chat_scheduleMessage(
            channel=user_id,
            text=message,
            blocks=message_blocks,
            post_at=post_at
        )
        print(f"[INFO] Scheduled direct message to Slack: {res['message']['text']}")
        return res.data['scheduled_message_id']
    except SlackApiError as e:
        print(f"[FAILED] Could not schedule direct message to Slack: {e.response['error']}")
        return None


AUTH_FAILURE_REASONS = {
    AuthenticationError.INVALID_CREDENTIALS: "Ugyldig brukernavn eller passord :key:",
    AuthenticationError.AUTH_TEMPORARILY_BLOCKED: "Midlertidig utestengt :no_entry:",
    AuthenticationError.TOKEN_EXTRACTION_FAILED: "Klarte ikke å hente autentiseringsnøkkel :sleuth_or_spy:",
    AuthenticationError.TOKEN_VALIDATION_FAILED: "Klarte ikke å verifisere autentiseringsnøkkel :interrobang:"
}


def notify_auth_failure_slack(slack_token: str, channel: str, user_id: str, error: AuthenticationError = None,
                              thread_ts: Optional[str] = None, check_run: bool = False) -> None:
    message = f"{':warning: Forhåndssjekk feilet!' if check_run else ':dizzy_face:'} Klarte ikke å logge inn som " \
              f"<@{user_id}>{f'. *{AUTH_FAILURE_REASONS[error]}*' if error in AUTH_FAILURE_REASONS else ''}"
    print(f"[INFO] Posting auth {'check ' if check_run else ''}failure notification to Slack")
    if not notify_slack(slack_token, channel, message, thread_ts):
        print(f"[FAILED] Could not post auth {'check ' if check_run else ''}failure notification to Slack")
        return
    print(f"[INFO] Auth {'check ' if check_run else ''}failure notification posted successfully to Slack.")
    return


BOOKING_FAILURE_REASONS = {
    BookingError.CLASS_MISSING: "Fant ikke timen :sleuth_or_spy:",
    BookingError.INCORRECT_START_TIME: "Feil starttid :clock7:",
    BookingError.MISSING_SCHEDULE_DAY: "Fant ikke riktig dag :calendar::mag:",
    BookingError.TOO_LONG_WAITING_TIME: "Ventetid før booking var for lang :sleeping:",
    BookingError.INVALID_CONFIG: "Ugyldig bookingkonfigurasjon :broken_heart:"
}


def notify_booking_failure_slack(slack_token: str, channel: str, user_id: str,
                                 _class_config: Class, error: BookingError = None,
                                 check_run: bool = False) -> None:
    class_name = f"{_class_config.display_name if _class_config.display_name is not None else _class_config.activity}"
    class_time = f"{WEEKDAYS[_class_config.weekday].lower()} " \
                 f"{datetime.time(_class_config.time.hour, _class_config.time.minute).strftime('%H:%M')}"
    msg = f"{':warning: Forhåndssjekk feilet! Kan ikke booke' if check_run else ':dizzy_face: Klarte ikke å booke'} " \
          f"*{class_name}* ({class_time}) for <@{user_id}>" \
          f"{f'. *{BOOKING_FAILURE_REASONS[error]}*' if error in BOOKING_FAILURE_REASONS else ''}"
    print(f"[INFO] Posting booking failure notification to Slack")
    if not notify_slack(slack_token, channel, msg):
        print(f"[FAILED] Could not post booking failure notification to Slack")
        return
    print(f"[INFO] Booking failure notification posted successfully to Slack.")
    return


WORKING_EMOJI_NAME = "sit-rezervo-working"


def notify_working_slack(slack_token: str, channel: str, message_ts: str):
    try:
        SlackClient(token=slack_token).reactions_add(
            channel=channel,
            timestamp=message_ts,
            name=WORKING_EMOJI_NAME
        )
        print(f"[INFO] 'Working' reaction posted successfully to Slack.")
    except SlackApiError as e:
        print(f"[FAILED] Could not post 'working' reaction to Slack: {e.response['error']}")


def notify_not_working_slack(slack_token: str, channel: str, message_ts: str):
    try:
        SlackClient(token=slack_token).reactions_remove(
            channel=channel,
            timestamp=message_ts,
            name=WORKING_EMOJI_NAME
        )
        print(f"[INFO] 'Working' reaction removed successfully from Slack message.")
    except SlackApiError as e:
        print(f"[FAILED] Could not remove 'working' reaction from Slack message: {e.response['error']}")


def notify_cancellation_slack(slack_token: str, channel: str,
                              source_ts: str, response_url: str):
    try:
        # Retrieve original message
        message = SlackClient(token=slack_token).conversations_history(
            channel=channel,
            latest=source_ts,
            limit=1,
            inclusive=True
        ).data["messages"][0]
        # Remove booking emoji and strike out original text
        new_message = f'~{message["blocks"][0]["text"]["text"]}~'.replace(BOOKING_EMOJI, CANCELLATION_EMOJI, 1)
        message["blocks"][0]["text"]["text"] = new_message
        # Remove all message actions
        message["blocks"] = [b for b in message["blocks"] if b["type"] != "actions"]
        # Update original message to reflect cancellation
        SlackWebhookClient(response_url).send(
            text=f'[AVBESTILT!] {message["text"]}',  # fallback text if 'blocks' don't work
            blocks=message["blocks"],
            replace_original=True
        )
        notify_not_working_slack(slack_token, channel, source_ts)
        # Notify in thread
        notify_slack(slack_token, channel, ":broken_heart: Timen er avbestilt", thread_ts=source_ts)
    except SlackApiError as e:
        print(f"[FAILED] Could not post cancellation notification to Slack: {e.response['error']}")
        return False
    return True


CANCELLATION_FAILURE_REASONS = {
    BookingError.CLASS_MISSING: "Klarte ikke å hente timen :sleuth_or_spy:",
    BookingError.CANCELLING_WITHOUT_BOOKING: "Timen var ikke booket :interrobang:"
}


def notify_cancellation_failure_slack(slack_token: str, channel: str, source_ts: str,
                                      error: BookingError | AuthenticationError) -> None:
    try:
        notify_not_working_slack(slack_token, channel, source_ts)
        # Notify failure in thread
        message = f":dizzy_face: Avbestilling feilet" \
                  f"{f'. {CANCELLATION_FAILURE_REASONS[error]}' if error in CANCELLATION_FAILURE_REASONS else ''}" \
                  f"{f'. {AUTH_FAILURE_REASONS[error]}' if error in AUTH_FAILURE_REASONS else ''}"
        notify_slack(slack_token, channel, message, thread_ts=source_ts)
    except SlackApiError as e:
        print(f"[FAILED] Could not post cancellation failure notification to Slack: {e.response['error']}")
        return
    print(f"[INFO] Cancellation failure notification posted successfully to Slack.")


def show_unauthorized_action_modal_slack(slack_token: str, trigger_id: str):
    try:
        SlackClient(token=slack_token).views_open(
            trigger_id=trigger_id,
            view={
                "type": "modal",
                "title": {
                    "type": "plain_text",
                    "text": "sit-rezervo"
                },
                "close": {
                    "type": "plain_text",
                    "text": ":cry: Lukk",
                    "emoji": True
                },
                "blocks": [
                    {
                        "type": "header",
                        "text": {
                            "type": "plain_text",
                            "text": ":four::zero::three:",
                            "emoji": True
                        }
                    },
                    {
                        "type": "section",
                        "fields": [
                            {
                                "type": "mrkdwn",
                                "text": "Det er lov å prøve seg :peach:"
                            }
                        ]
                    }
                ]
            }
        )
    except SlackApiError as e:
        print(f"[FAILED] Could not display unauthorized action modal in Slack: {e.response['error']}")
        return
    print(f"[INFO] Unauthorized action modal displayed successfully in Slack.")


def notify_booking_slack(slack_token: str, channel: str, user_id: str, booked_class: Dict[str, Any], ical_url: str,
                         transfersh_url: Optional[str], scheduled_reminder_id: Optional[str]) -> None:
    message_blocks = booking_message_blocks(booked_class, user_id, scheduled_reminder_id)
    if transfersh_url:
        filename = f"{booked_class['id']}.ics"
        print(f"[INFO] Uploading {filename} to {transfersh_url}")
        try:
            ical_tsh_url = upload_ical_to_transfersh(transfersh_url, ical_url, filename)
            message_blocks = booking_message_blocks(booked_class, user_id, ical_tsh_url, scheduled_reminder_id)
        except RequestException:
            print(f"[WARNING] Could not upload ical event to transfer.sh instance, skipping ical link in notification.")
    print(f"[INFO] Posting booking notification to Slack")
    if not notify_slack(slack_token, channel, message_blocks['message'], message_blocks['blocks']):
        print(f"[FAILED] Could not post booking notification to Slack")
        return
    print(f"[INFO] Booking notification posted successfully to Slack.")
    return


BOOKING_EMOJI = ":robot_face:"
CANCELLATION_EMOJI = ":no_entry:"


def booking_message_blocks(booked_class: Dict[str, Any], user_id: str, ical_tsh_url: Optional[str] = None,
                           scheduled_reminder_id: Optional[str] = None):
    buttons = [
        {
            "type": "button",
            "action_id": SLACK_ACTION_CANCEL_BOOKING,
            "value": json.dumps(
                {
                    "userId": user_id,
                    "classId": str(booked_class['id']),
                    "scheduledReminderMessageId": scheduled_reminder_id
                }
            ),
            "text": {
                "type": "plain_text",
                "text": ":no_entry: Avbestill"
            },
            "confirm": {
                "title": {
                    "type": "plain_text",
                    "text": "Er du sikker?"
                },
                "text": {
                    "type": "plain_text",
                    "text": f"Du er i ferd med å avbestille {booked_class['name']} ({booked_class['from']}). "
                            f"Dette kan ikke angres!"
                },
                "confirm": {
                    "type": "plain_text",
                    "text": "Avbestill"
                },
                "deny": {
                    "type": "plain_text",
                    "text": "Avbryt"
                },
                "style": "danger"
            }
        }
    ]
    if ical_tsh_url:
        buttons.insert(0, {
            "type": "button",
            "action_id": SLACK_ACTION_ADD_BOOKING_TO_CALENDAR,
            "text": {
                "type": "plain_text",
                "text": ":calendar: Legg i kalender"
            },
            "url": ical_tsh_url
        })
    message = f"{BOOKING_EMOJI} *{booked_class['name']}* ({booked_class['from']}) er booket for <@{user_id}>"
    blocks = [
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": message
            }
        }
    ]
    if len(buttons) > 0:
        blocks.append({
            "type": "actions",
            "elements": buttons
        })
    return {
        "message": message,
        "blocks": blocks
    }


def schedule_class_reminder_slack(slack_token: str, user_id: str, _class: Dict[str, Any], hours_before: int) \
        -> Optional[str]:
    start_time = datetime.datetime.fromisoformat(_class['from'])
    reminder_time = start_time - datetime.timedelta(hours=hours_before)
    reminder_timestamp = int(time.mktime(reminder_time.timetuple()))
    message = f"Husk *{_class['name']}* ({_class['from']}) om {hours_before} timer!"
    return schedule_dm_slack(slack_token, user_id, reminder_timestamp, message)



import re
from enum import Enum, auto
from typing import Union

import requests
from requests import Session

from .consts import AUTH_URL, BOOKING_URL, TOKEN_VALIDATION_URL

USER_AGENT = "Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:100.0) Gecko/20100101 Firefox/100.0"


class AuthenticationError(Enum):
    ERROR = auto()
    TOKEN_EXTRACTION_FAILED = auto()
    TOKEN_VALIDATION_FAILED = auto()
    AUTH_TEMPORARILY_BLOCKED = auto()
    INVALID_CREDENTIALS = auto()


def authenticate(email: str, password: str) -> Union[str, AuthenticationError]:
    session = Session()
    user_agent_header = {'User-Agent': USER_AGENT}
    auth_res = session.post(AUTH_URL, {
        "name": email,
        "pass": password,
        "form_id": "user_login"
    }, headers=user_agent_header)
    auth_soup = re.sub(" +", " ", auth_res.text.replace("\n", ""))
    login_blocked_matches = re.search(r"Feilmelding.*?midlertidig blokkert", auth_soup)
    if login_blocked_matches is not None:
        print("[ERROR] Authentication failed, authentication temporarily blocked")
        return AuthenticationError.AUTH_TEMPORARILY_BLOCKED
    invalid_credentials_matches = re.search(r"Feilmelding.*?ukjent brukernavn eller passord", auth_soup)
    if invalid_credentials_matches is not None:
        print("[ERROR] Authentication failed, invalid credentials")
        return AuthenticationError.INVALID_CREDENTIALS
    booking_res = session.get(BOOKING_URL, headers=user_agent_header)
    booking_soup = re.sub(" +", " ", booking_res.text.replace("\n", ""))
    cdata_token_matches = re.search(r"<!\[CDATA\[.*?iBookingPreload\(.*?token:.*?\"(.+?)\".*?]]>", booking_soup)
    if cdata_token_matches is None:
        print("[ERROR] Failed to extract authentication token!")
        return AuthenticationError.TOKEN_EXTRACTION_FAILED
    try:
        token = cdata_token_matches.group(1)
    except IndexError:
        print("[ERROR] Failed to extract authentication token!")
        return AuthenticationError.TOKEN_EXTRACTION_FAILED
    # Validate token
    token_validation = session.post(TOKEN_VALIDATION_URL, {"token": token})
    if token_validation.status_code != requests.codes.OK:
        print("[ERROR] Validation of authentication token failed")
        return AuthenticationError.TOKEN_VALIDATION_FAILED
    token_info = token_validation.json()
    if 'info' in token_info and token_info['info'] == "client-readonly":
        print("[ERROR] Authentication failed, only acquired public readonly access")
        return AuthenticationError.ERROR
    user = token_info['user']
    print(f"[INFO] Authenticated as {user['firstname']} {user['lastname']} ({user['email']})")
    return token



import time
from typing import Dict, Any, Optional, Union

from .auth import AuthenticationError, authenticate
from .booking import book_class, cancel_booking
from .config import Notifications
from .consts import ICAL_URL
from .errors import BookingError
from .notify.notify import notify_booking


def try_book_class(token: str, _class: Dict[str, Any], max_attempts: int,
                   notifications_config: Optional[Notifications] = None) -> Optional[BookingError]:
    if max_attempts < 1:
        print(f"[ERROR] Max booking attempts should be a positive number")
        return BookingError.INVALID_CONFIG
    booked = False
    attempts = 0
    while not booked:
        booked = book_class(token, _class['id'])
        attempts += 1
        if booked:
            break
        if attempts >= max_attempts:
            break
        sleep_seconds = 2 ** attempts
        print(f"[INFO] Exponential backoff, retrying in {sleep_seconds} seconds...")
        time.sleep(sleep_seconds)
    if not booked:
        print(f"[ERROR] Booking failed after {attempts} attempt" + ("s" if attempts != 1 else ""))
        return BookingError.ERROR
    print(f"[INFO] Successfully booked class" + (f" after {attempts} attempts!" if attempts != 1 else "!"))
    if notifications_config:
        ical_url = f"{ICAL_URL}/?id={_class['id']}&token={token}"
        notify_booking(notifications_config, _class, ical_url)
    return None


def try_cancel_booking(token: str, _class: Dict[str, Any], max_attempts: int) -> Optional[BookingError]:
    if _class["userStatus"] != "booked":
        print(f"[ERROR] Class is not booked, cancellation is not possible")
        return BookingError.CANCELLING_WITHOUT_BOOKING
    if max_attempts < 1:
        print(f"[ERROR] Max booking cancellation attempts should be a positive number")
        return BookingError.INVALID_CONFIG
    cancelled = False
    attempts = 0
    while not cancelled:
        cancelled = cancel_booking(token, _class['id'])
        attempts += 1
        if cancelled:
            break
        if attempts >= max_attempts:
            break
        sleep_seconds = 2 ** attempts
        print(f"[INFO] Exponential backoff, retrying in {sleep_seconds} seconds...")
        time.sleep(sleep_seconds)
    if not cancelled:
        print(f"[ERROR] Booking cancellation failed after {attempts} attempt" + ("s" if attempts != 1 else ""))
        return BookingError.ERROR
    print(f"[INFO] Successfully cancelled booking" + (f" after {attempts} attempts!" if attempts != 1 else "!"))
    return None


def try_authenticate(email: str, password: str, max_attempts: int) -> Union[str, AuthenticationError]:
    if max_attempts < 1:
        return AuthenticationError.ERROR
    success = False
    attempts = 0
    result = None
    while not success:
        result = authenticate(email, password)
        success = not isinstance(result, AuthenticationError)
        attempts += 1
        if success:
            break
        if result == AuthenticationError.INVALID_CREDENTIALS:
            print("[ERROR] Invalid credentials, aborting authentication to avoid lockout")
            break
        if result == AuthenticationError.AUTH_TEMPORARILY_BLOCKED:
            print("[ERROR] Authentication temporarily blocked, aborting")
            break
        if attempts >= max_attempts:
            break
        sleep_seconds = 2 ** attempts
        print(f"[INFO] Exponential backoff, retrying in {sleep_seconds} seconds...")
        time.sleep(sleep_seconds)
    if not success:
        print(f"[ERROR] Authentication failed after {attempts} attempt" + ("s" if attempts != 1 else ""))
    return result if result is not None else AuthenticationError.ERROR



import time
from datetime import datetime
from typing import Optional

import typer
import uvicorn
from pytz import timezone

from . import api
from .auth import AuthenticationError
from .booking import find_class
from .config import config_from_stream
from .cron_generator import generate_booking_cron_job
from .errors import BookingError
from .main import try_book_class, try_authenticate
from .notify.notify import notify_booking_failure, notify_auth_failure
from .utils.time_utils import readable_seconds
from .api import api as api_app, get_configs

cli = typer.Typer()


@cli.command()
def book(
        class_id: int,
        check_run: bool = typer.Option(False, "--check", help="Perform a dry-run to verify that booking is possible"),
        config_stream: typer.FileText = typer.Option(
            "config.yaml",
            "--config-file", "-c",
            encoding="utf-8",
            help="Configurations file"
        ),
) -> None:
    """
    Book the class with config index matching the given class id
    """
    print("[INFO] Loading config...")
    config = config_from_stream(config_stream)
    if config is None:
        print("[ERROR] Failed to load config, aborted.")
        return
    if not 0 <= class_id < len(config.classes):
        print(f"[ERROR] Class index out of bounds")
        return
    _class_config = config.classes[class_id]
    if config.booking.max_attempts < 1:
        print(f"[ERROR] Max booking attempts should be a positive number")
        if config.notifications is not None:
            notify_booking_failure(config.notifications, _class_config, BookingError.INVALID_CONFIG, check_run)
        return
    print("[INFO] Authenticating...")
    auth_result = try_authenticate(config.auth.email, config.auth.password, config.auth.max_attempts)
    if isinstance(auth_result, AuthenticationError):
        print("[ERROR] Abort!")
        if config.notifications is not None:
            notify_auth_failure(config.notifications, auth_result, check_run)
        return
    class_search_result = find_class(auth_result, _class_config)
    if isinstance(class_search_result, BookingError):
        print("[ERROR] Abort!")
        if config.notifications is not None:
            notify_booking_failure(config.notifications, _class_config, class_search_result, check_run)
        return
    if check_run:
        print("[INFO] Check complete, all seems fine.")
        raise typer.Exit()
    _class = class_search_result
    if _class['bookable']:
        print("[INFO] Booking is already open, booking now!")
    else:
        # Retrieve booking opening, and make sure it's timezone aware
        tz = timezone(config.booking.timezone)
        opening_time = tz.localize(datetime.fromisoformat(_class['bookingOpensAt']))
        timedelta = opening_time - datetime.now(tz)
        wait_time = timedelta.total_seconds()
        wait_time_string = readable_seconds(wait_time)
        if wait_time > config.booking.max_waiting_minutes * 60:
            print(f"[ERROR] Booking waiting time was {wait_time_string}, "
                  f"but max is {config.booking.max_waiting_minutes} minutes. Aborting.")
            if config.notifications is not None:
                notify_booking_failure(config.notifications, _class_config, BookingError.TOO_LONG_WAITING_TIME)
            raise typer.Exit(1)
        print(f"[INFO] Scheduling booking at {datetime.now(tz) + timedelta} "
              f"(about {wait_time_string} from now)")
        time.sleep(wait_time)
        print(f"[INFO] Awoke at {datetime.now(tz)}")
    booking_result = try_book_class(auth_result, _class, config.booking.max_attempts, config.notifications)
    if isinstance(booking_result, BookingError):
        if config.notifications is not None:
            notify_booking_failure(config.notifications, _class_config, booking_result, check_run)
        raise typer.Exit(1)


@cli.command()
def cron(
        config_streams: list[typer.FileText] = typer.Option(
            ["config.yaml"],
            "--config-file", "-c",
            encoding="utf-8",
            help="Configurations file (multiple allowed)"
        ),
        output_file: Optional[typer.FileTextWrite] = typer.Option(None, "--output-file", "-o", encoding="utf-8"),
):
    """
    Generate cron jobs for class booking
    """
    cron_spec = ""
    for config_stream in config_streams:
        config = config_from_stream(config_stream)
        for i, c in enumerate(config.classes):
            cron_spec += generate_booking_cron_job(i, c, config.cron, config_stream.name)
    if output_file is not None:
        output_file.write(cron_spec + "\n")
        raise typer.Exit()
    print(cron_spec)


@cli.command(
    name="api",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}  # Enabled to support uvicorn options
)
def serve_api(
        ctx: typer.Context,
        config_streams: list[typer.FileText] = typer.Option(
            ["config.yaml"],
            "--config-file", "-c",
            encoding="utf-8",
            help="Configurations file (multiple allowed)"
        ),
):
    """
    Start a web server to handle Slack message interactions

    Actually a wrapper around uvicorn, and supports passing additional options to the underlying uvicorn.run() command.
    """
    ctx.args.insert(0, f"{api.__name__}:api")
    configs = [config_from_stream(cs) for cs in config_streams]
    api_app.dependency_overrides[get_configs] = lambda: configs
    uvicorn.main.main(args=ctx.args)


@cli.callback()
def callback():
    """
    Automatic booking of Sit Trening group classes
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sit_rezervo', 'sit_rezervo.notify', 'sit_rezervo.utils']

package_data = \
{'': ['*']}

install_requires = \
['dataclass-wizard[yaml]>=0.22.2,<0.23.0',
 'fastapi>=0.88.0,<0.89.0',
 'pydantic>=1.10.2,<2.0.0',
 'python-multipart>=0.0.5,<0.0.6',
 'pytz>=2022.7,<2023.0',
 'requests>=2.28.1,<3.0.0',
 'slack-sdk>=3.19.5,<4.0.0',
 'typer>=0.7.0,<0.8.0',
 'uvicorn>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['sit-rezervo = sit_rezervo.cli:cli']}

setup_kwargs = {
    'name': 'sit-rezervo',
    'version': '0.1.0',
    'description': 'Automatic booking of Sit Trening group classes',
    'long_description': '# sit-rezervo\n\n[![PyPI](https://img.shields.io/pypi/v/sit-rezervo)](https://pypi.org/project/sit-rezervo/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/sit-rezervo)\n\nAutomatic booking of [Sit Trening group classes](https://www.sit.no/trening/gruppe)',
    'author': 'Mathias Oterhals Myklebust',
    'author_email': 'mathias@oterbust.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mathiazom/sit-rezervo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
